var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
const EMPTY_STACK_TRACE = "<stack trace empty>";
export class EError extends Error {
    constructor(message, causeOrOptions, info) {
        var _a, _b;
        let _cause = undefined;
        let _private = undefined;
        let _info = undefined;
        if (causeOrOptions instanceof Error) {
            _cause = causeOrOptions;
            if (info) {
                const { cause, private: isPrivate } = info, rest = __rest(info, ["cause", "private"]);
                _private = isPrivate;
                _info = rest;
            }
        }
        else if (causeOrOptions) {
            const { cause, private: isPrivate } = causeOrOptions, rest = __rest(causeOrOptions, ["cause", "private"]);
            _cause = cause;
            _private = isPrivate;
            _info = rest;
        }
        if (_private || !_cause) {
            super(message);
        }
        else {
            super(`${message}: ${_cause.message}`);
            this.summary = message;
        }
        if (_private)
            this.private = true;
        if (_cause)
            this.cause = _cause;
        if (_info && Object.keys(_info).length > 0)
            this.info = _info;
        Object.setPrototypeOf(this, new.target.prototype);
        this.name = new.target.name;
        (_b = (_a = Error).captureStackTrace) === null || _b === void 0 ? void 0 : _b.call(_a, this, new.target);
    }
    static cause(error) {
        const cause = error.cause;
        if (cause instanceof Error)
            return cause;
        return null;
    }
    static info(error) {
        var _a;
        return (_a = error.info) !== null && _a !== void 0 ? _a : null;
    }
    static findCause(error, ...names) {
        const causeNames = new Set(names);
        let errorCause = error;
        while (errorCause) {
            if (causeNames.has(errorCause.name))
                return errorCause;
            errorCause = EError.cause(errorCause);
        }
        return null;
    }
    findCause(...names) {
        return EError.findCause(this, ...names);
    }
    static findCauseIf(error, predicate) {
        let errorCause = error;
        while (errorCause) {
            if (predicate(error))
                return errorCause;
            errorCause = EError.cause(errorCause);
        }
        return null;
    }
    findCauseIf(predicate) {
        return EError.findCauseIf(this, predicate);
    }
    static findCauses(error, ...names) {
        const causeNames = new Set(names);
        const causes = [];
        let errorCause = error;
        while (errorCause) {
            if (causeNames.has(errorCause.name))
                causes.push(errorCause);
            errorCause = EError.cause(errorCause);
        }
        return causes;
    }
    findCauses(...names) {
        return EError.findCauses(this, ...names);
    }
    static findCausesIf(error, predicate) {
        const causes = [];
        let errorCause = error;
        while (errorCause) {
            if (predicate(error))
                causes.push(error);
            errorCause = EError.cause(errorCause);
        }
        return causes;
    }
    findCausesIf(predicate) {
        return EError.findCausesIf(this, predicate);
    }
    static hasCause(error, ...names) {
        return EError.findCause(error, ...names) !== null;
    }
    hasCause(...names) {
        return EError.hasCause(this, ...names);
    }
    static fullStack(error) {
        var _a, _b;
        const causeError = EError.cause(error);
        if (causeError)
            return (((_a = error.stack) !== null && _a !== void 0 ? _a : EMPTY_STACK_TRACE) +
                "\n\n" +
                `caused by: ${EError.fullStack(causeError)}`);
        return (_b = error.stack) !== null && _b !== void 0 ? _b : EMPTY_STACK_TRACE;
    }
    fullStack() {
        return EError.fullStack(this);
    }
    static toJSON(error, options = {}) {
        const { stack, shallow } = options;
        const json = {
            name: error.name,
            message: error.message,
        };
        if (error instanceof EError && error.summary) {
            json.summary = error.summary;
        }
        if (!shallow) {
            const errorCause = EError.cause(error);
            if (errorCause) {
                json.cause = EError.toJSON(errorCause, options);
            }
        }
        const errorInfo = EError.info(error);
        if (errorInfo) {
            json.info = errorInfo;
        }
        if (stack) {
            const stack = error.stack;
            if (stack)
                json.stack = stack.split("\n");
        }
        return json;
    }
    toJSON(options = {}) {
        return EError.toJSON(this, options);
    }
}
